==> cnn2snn-2.3.3.tar.gz: cnn2snn-2.3.3/cnn2snn/quantizeml/depthwise_conv2d_transpose.py <==
#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDepthwiseConv2DTranspose to Akida.
"""
import numpy as np

from quantizeml.layers import (QuantizedDepthwiseConv2DTranspose, QuantizedReLU,
                               WeightQuantizer, AlignedWeightQuantizer)
from akida import DepthwiseConv2DTranspose

from ..akida_versions import AkidaVersion, get_akida_version
from .activations import set_relu_variables
from .weights import broadcast_and_set_variable
from .outputs import set_output_variables
from .blocks import parse_block_additional_layers, get_block_out_quantizer


def _set_depthwise_conv2d_transpose_variables(layer_ak, layer_k):
    """Computes and sets the variables for an Akida DepthwiseConv2DTranspose layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(layer_k, QuantizedDepthwiseConv2DTranspose)
    assert isinstance(layer_k.weight_quantizer, WeightQuantizer)
    if layer_k.use_bias:
        assert isinstance(layer_k.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = layer_k.weight_quantizer.qweights.value.fp.values.numpy()
    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if layer_k.use_bias:
        bias_quantizer = layer_k.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)

        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)

    # Set input shift
    input_shift = getattr(layer_k, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                   layer_k.input_shift.value.numpy().astype(np.uint8))


def _parse_depthwise_conv2d_transpose(layer):
    """Parses a quantizeml.QuantizedDepthwiseConv2DTranspose parameters.

    Args:
        layer (:obj:`tf.keras.Layer`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedDepthwiseConv2DTranspose)

    # Make sure the DepthwiseConv2D kernel size and stride params are square
    assert layer.kernel_size[0] == layer.kernel_size[1], (
        "DepthwiseConv2DTranspose kernel_size should be square")
    assert layer.strides[0] == layer.strides[1], (
        "DepthwiseConv2DTranspose strides should be the same on both dimension")
    assert layer.kernel_size[0] in [3, 4], (
        "DepthwiseConv2DTranspose supports kernel_size of 3 or 4 only")
    assert layer.strides[0] == 2, "DepthwiseConv2DTranspose supports stride 2 only"

    # Padding value must be same
    assert layer.padding == "same", "DepthwiseConv2DTranspose handles only 'same' padding"

    # The only weight bitwidth supported is [4, 8]
    weight_bits = layer.weight_quantizer.bitwidth
    assert weight_bits in [4, 8]

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1

    return dict(
        kernel_size=layer.kernel_size[0],
        buffer_bits=buffer_bits,
        activation=False,
        name=layer.name
    )


def convert_depthwise_conv2d_transpose_block(model_ak, layers):
    """Converts a depthwise convolutional block into an akida DepthwiseConv2DTranspose layer.

    The expected sequence is:

    - QuantizedDepthwiseConv2DTranspose,
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        layers (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        int: the number of layers in the block or 0 if the first layer is not a
        QuantizedDepthwiseConv2DTranspose.
    """
    if len(layers) == 0 or not isinstance(layers[0], QuantizedDepthwiseConv2DTranspose):
        return 0

    if get_akida_version() == AkidaVersion.v1:
        return 0

    dw_conv = layers[0]

    # Evaluate the depthwise convolutional transpose layer parameters
    conv_params = _parse_depthwise_conv2d_transpose(dw_conv)

    # Identify the next layers
    next_layers = parse_block_additional_layers(layers, conv_params)

    layer_block = [dw_conv] + list(next_layers.values())

    # Get block output quantizer
    out_quantizer = get_block_out_quantizer(layer_block)

    # add output quantizer bitwidth parameter
    if out_quantizer:
        conv_params["output_bits"] = out_quantizer.bitwidth
    else:
        conv_params["output_bits"] = conv_params["buffer_bits"]

    # Create Akida layer
    dw_conv_ak = DepthwiseConv2DTranspose(**conv_params)
    # Add layer to the model to build its internal variables
    model_ak.add(dw_conv_ak)

    # Set base variables
    _set_depthwise_conv2d_transpose_variables(dw_conv_ak, dw_conv)

    # Set optional activation variables if we have a ReLU
    relu_layer = next_layers.get(QuantizedReLU, None)
    if relu_layer:
        set_relu_variables(dw_conv_ak, relu_layer)

    # Set optional output_quantizer variables
    if out_quantizer:
        set_output_variables(dw_conv_ak, out_quantizer)

    return len(layer_block)

==> cnn2snn-2.3.7.tar.gz: cnn2snn-2.3.7/cnn2snn/quantizeml/depthwise_conv2d_transpose.py <==
#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDepthwiseConv2DTranspose to Akida.
"""
import numpy as np

from quantizeml.layers import (QuantizedDepthwiseConv2DTranspose, QuantizedReLU,
                               WeightQuantizer, AlignedWeightQuantizer)
from akida import DepthwiseConv2DTranspose

from ..akida_versions import AkidaVersion, get_akida_version
from .activations import set_relu_variables
from .weights import broadcast_and_set_variable
from .outputs import set_output_v2_variables
from .blocks import parse_block_additional_layers, get_block_out_quantizer


def _set_depthwise_conv2d_transpose_variables(layer_ak, layer_k):
    """Computes and sets the variables for an Akida DepthwiseConv2DTranspose layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(layer_k, QuantizedDepthwiseConv2DTranspose)
    assert isinstance(layer_k.weight_quantizer, WeightQuantizer)
    if layer_k.use_bias:
        assert isinstance(layer_k.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = layer_k.weight_quantizer.qweights.value.fp.values.numpy()
    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if layer_k.use_bias:
        bias_quantizer = layer_k.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)

        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)

    # Set input shift
    input_shift = getattr(layer_k, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                   layer_k.input_shift.value.numpy().astype(np.uint8))


def _parse_depthwise_conv2d_transpose(layer):
    """Parses a quantizeml.QuantizedDepthwiseConv2DTranspose parameters.

    Args:
        layer (:obj:`tf.keras.Layer`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedDepthwiseConv2DTranspose)

    # Make sure the DepthwiseConv2D kernel size and stride params are square
    assert layer.kernel_size[0] == layer.kernel_size[1], (
        "DepthwiseConv2DTranspose kernel_size should be square")
    assert layer.strides[0] == layer.strides[1], (
        "DepthwiseConv2DTranspose strides should be the same on both dimension")
    assert layer.kernel_size[0] in [3, 4], (
        "DepthwiseConv2DTranspose supports kernel_size of 3 or 4 only")
    assert layer.strides[0] == 2, "DepthwiseConv2DTranspose supports stride 2 only"

    # Padding value must be same
    assert layer.padding == "same", "DepthwiseConv2DTranspose handles only 'same' padding"

    # The only weight bitwidth supported is [4, 8]
    weight_bits = layer.weight_quantizer.bitwidth
    assert weight_bits in [4, 8]

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1

    return dict(
        kernel_size=layer.kernel_size[0],
        buffer_bits=buffer_bits,
        activation=False,
        name=layer.name
    )


def convert_depthwise_conv2d_transpose_block(model_ak, layers):
    """Converts a depthwise convolutional block into an akida DepthwiseConv2DTranspose layer.

    The expected sequence is:

    - QuantizedDepthwiseConv2DTranspose,
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        layers (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        int: the number of layers in the block or 0 if the first layer is not a
        QuantizedDepthwiseConv2DTranspose.
    """
    if len(layers) == 0 or not isinstance(layers[0], QuantizedDepthwiseConv2DTranspose):
        return 0

    if get_akida_version() == AkidaVersion.v1:
        return 0

    dw_conv = layers[0]

    # Evaluate the depthwise convolutional transpose layer parameters
    conv_params = _parse_depthwise_conv2d_transpose(dw_conv)

    # Identify the next layers
    next_layers = parse_block_additional_layers(layers, conv_params)

    layer_block = [dw_conv] + list(next_layers.values())

    # Get block output quantizer
    out_quantizer = get_block_out_quantizer(layer_block)

    # add output quantizer bitwidth parameter
    if out_quantizer:
        conv_params["output_bits"] = out_quantizer.bitwidth
    else:
        conv_params["output_bits"] = conv_params["buffer_bits"]

    # Create Akida layer
    dw_conv_ak = DepthwiseConv2DTranspose(**conv_params)
    # Add layer to the model to build its internal variables
    model_ak.add(dw_conv_ak)

    # Set base variables
    _set_depthwise_conv2d_transpose_variables(dw_conv_ak, dw_conv)

    # Set optional activation variables if we have a ReLU
    relu_layer = next_layers.get(QuantizedReLU, None)
    if relu_layer:
        set_relu_variables(dw_conv_ak, relu_layer)

    # Set optional output_quantizer variables
    if out_quantizer:
        set_output_v2_variables(dw_conv_ak, out_quantizer)

    return len(layer_block)
